from types import GeneratorType

from context import PipelineContext
from utils import is_dag, topological_sort


class Pipeline(PipelineContext):
    def __init__(self, pipeline_name: str):
        super().__init__()
        self.name = pipeline_name

    def validate(self):
        if not is_dag(self.edges):
            raise ValueError("Pipeline contains a cycle")

    def run(self):
        self.validate()

        execution_order = topological_sort(self.edges)
        context = {}

        print(f"[RUNNING] Pipeline '{self.name}'")
        print(f"{execution_order=}")
        print(f"{self.nodes=}")
        print(f"{context=}")

        for op_id in execution_order:
            node = self.nodes[op_id]
            inputs = [context[parent.operator_id] for parent in getattr(node, 'upstream', [])]
            print(f"  - Executing operator: {op_id} with {inputs=}")
            result = node.process(*inputs)
            # If it's a generator, materialize it for memoization
            if isinstance(result, GeneratorType):
                result = list(result)
            context[op_id] = result
